/*
* Copyright (c) 1980 Regents of the University of California.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. All advertising materials mentioning features or use of this software
*    must display the following acknowledgement:
*	This product includes software developed by the University of
*	California, Berkeley and its contributors.
* 4. Neither the name of the University nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/

/*
* 1999-02-22 Arkadiusz Mi�kiewicz <misiek@pld.ORG.PL>
* - added Native Language Support
*
* 2000-07-30 Per Andreas Buer <per@linpro.no> - added "q"-option
*/

/*
* 2009-01-20 Ed Brand <ebrand@brandint.com>
*  - Modified to allow logging of keystrokes to a file
*  - This file is based off a  modified version of the script.c 
*    source code mentioned in the above Copyright
*/
#include <stdio.h>
#include <stdlib.h>
#include <paths.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <syslog.h>
#include <pwd.h>
#include <errno.h>
#include <locale.h>
#include "nls.h"
#include "config.h"

#ifdef __linux__
#include <unistd.h>
#include <string.h>
#endif 

#ifdef HAVE_LIBUTIL
#include <pty.h>
#endif

//Prototypes
void finish(int);
void done(void);
void fail(void);
void resize(int);
void fixtty(void);
void getmaster(void);
void getslave(void);
void doinput(void);
void dooutput(void);
void doshell(void);
void do_setuid(void);
void undo_setuid(void);
void do_setgid(void);
void undo_setgid(void);
int read_conf(const char *, char **);
int readConf(void);
int logCmd2(char *, char *, char *,char *, char *);
int createSlog(void);

char	*shell;
FILE	*fscript;
FILE    *pinput; // pipe command input
int	master;
int	slave;
int	child;
int	subchild;
char	*fname; 
char	*sfname; //short file name
char	*comment;
int commentLen=0; //length of comment after entering reason
struct  passwd *upasswd;
struct	termios tt;
struct	winsize win;
int	lb;
int	l;
#ifndef HAVE_LIBUTIL
char	line[] = "/dev/ptyXX";
#endif
char	*cflg = NULL;
int	debug=0;
int	fflg = 0;
int	hflg = 1;  //on by default
int	qflg = 0;
int	tflg = 0;
char *ruser;
static char *progname;

//defaults if sslogger.conf is foobar
int log_all_cmds=0;
int commentSize=100; //Max comment len //TODO: change name to max_comment to match config name
int minCommentSize=0; //Min comment len //TODO: change name to min_comment to match config name
int allow_user_replay=1; //allow user read their own logs

#define LINE_LEN 100
int read_conf( const char *key, char **value) {
    //reads config_file looking fir key, 
    // sets  value if found, null otherwise
    // returns -1 on FNF, 0 on keynot found, 1 on found
    FILE *fp;
    char line[LINE_LEN];
    int end;
    int keyLen=strlen(key);
    const char *conf_file=STRINGIFY(CONF_FILE);
    //if (*value!=NULL) free(*value);
    fp = fopen(conf_file, "r");    
    if (fp == NULL ) {
        fprintf ( stderr,"Can't open config file: %s\n",conf_file);
        return(-1); /* Can't open file */
    }
    
    while (fgets(line, LINE_LEN, fp)) {
        /* All options are key=value (no spaces)*/
        if (line == NULL) continue;        
        //if (debug) printf("line:%s",line);
        if (strncmp(line, key, keyLen) == 0) { //we have a match //we have a match
            end = strlen(line);
            if (line[end-1] == '\n')
                line[end-1] = 0; /* Remove trailing newline */
        *value = strdup(line+keyLen+1); //add equal to  key=val
        if (debug) printf("Found key:%s val:%s\n",key,*value);
        if (fp!=NULL) fclose(fp);
        return(1); //happy return
        }
    }
    return 0; //unhappy return
}

/* Remember the effective and real UIDs. */
static uid_t euid, ruid;
static gid_t geuid,gruid;

static void
die_if_link(char *fn) {
    struct stat s;
    if (lstat(fn, &s) == 0 && (S_ISLNK(s.st_mode) || s.st_nlink > 1)) {
        fprintf(stderr,
            _("Warning: `%s' is a link.\n"
            "Use `%s [options] %s' if you really "
            "want to use it.\n"
            "Sslogger not started.\n"),
            fn, progname, fn);
        exit(1);
    }
}

/*
* Stop extremely silly gcc complaint on %c:
*  warning: `%c' yields only last 2 digits of year in some locales
*/
static void
my_strftime(char *buf, size_t len, const char *fmt, const struct tm *tm) {
    strftime(buf, len, fmt, tm);
}

int logCmd2(char *user, char *asUser, char *cmd, char *keyLogfname, char *reason) {
    char *message;
    char *logfname;
    const char *log_dir=STRINGIFY(LOG_DIR);
    const char *log_file=STRINGIFY(LOGFILE);
    asprintf(&logfname,"%s/%s",log_dir,log_file);
    do_setuid();
    do_setgid();
    FILE *log;
    log=fopen(logfname,"a");
    
    if (log == NULL ) {
        /* Can't open file */
        if (debug) {
            printf ("set euid: %d\n",euid); //suid user
            printf ("real ruid: %d\n",ruid); //real user
        }
        undo_setuid();
        undo_setgid();
        fprintf ( stderr,"Error: can't open logfile: %s\n",logfname);
        sleep(1);
        return (0);
        
    }

    /*time format for logfname*/
    char outstr[200];
    time_t t;
    struct tm *tmp;
    t = time(NULL);
    tmp = localtime(&t);
    if (tmp == NULL) {
        perror("localtime");
        exit(EXIT_FAILURE);
    }
    if (strftime(outstr, sizeof(outstr), "%F %H:%M:%S", tmp) == 0) {
        fprintf(stderr, "strftime returned 0");
        undo_setuid();
        undo_setgid();
        exit(EXIT_FAILURE);
    } 
  
    if (debug) printf("opening log file %s\n",logfname);
    int p=getpid(); //get the pid
    fprintf(log,"%s %s[%d];",outstr,progname,p);
    //fprintf(log," user:%s; as:%s; %s; logfile:%s; reason:%s\n",user,asUser,cmd,keyLogfname,reason);
    fprintf(log," user:%s; as:%s; %s; logfile:%s; reason:%s\n",user,asUser,cmd,fname,reason);
    fclose(log);
 
    if ((!cflg) && (strlen(reason) > commentSize) && (minCommentSize>0) ) { 
        /* cflag - command to run (non interactive)
        *  minCommentSize - the minimum comment len, if zero, there is no comment "(null)"
        *  commentSize - maximum comment len for syslog comments, if zero no comment is required
        *  reason - is the comment, (null) if commentSize=0
        */
        if (debug) {
            printf("whacking comment len to:%d\n",commentSize);
            printf("reason-length:%d syslogCommentSize:%d minCommentSize:%d\n",strlen(reason),commentSize,minCommentSize);
        }
        reason[commentSize]='\0';
    }

    asprintf(&message,"user:%s; as:%s; %s; logfile:%s; reason:%s\n",user,asUser,cmd,keyLogfname,reason);
    //openlog (progname, LOG_PID,LOG_AUTHPRIV);
    syslog(LOG_INFO,message);
    //closelog();
    undo_setuid();
    undo_setgid();
    return 1; //happy ending
}


/* Read slogger.conf an set global vars */
int readConf (void) {
    commentSize=MAXCOMMENTSIZE; //set default size
    minCommentSize=MINCOMMENTSIZE; //dito
    /* Get options from config file */
    char *value;
    int found=0;
    found=read_conf("log_all_cmds",&value);
    if (found==1) {
        //printf ("found, result is: %s\n",value);
        log_all_cmds=atoi(value);
    }
    else {
        //printf ("Not found\n");
        log_all_cmds=0; // set to not log all cmds
    }
    found=read_conf("max_comment",&value);
    if (found==1) {
        //printf ("found, result is: %s\n",value);
        commentSize=atoi(value);
    }
    
    found=read_conf("min_comment",&value);
    if (found==1) {
          //printf ("found, result is: %s\n",value);
        minCommentSize=atoi(value);
    }

    found=read_conf("allow_user_replay",&value);
    if (found==1) {
          //printf ("found, result is: %s\n",value);
        allow_user_replay=atoi(value);
    }

    if (debug) printf("log_all_cmds=%d\n",log_all_cmds);
    if (debug) printf("Max commentSize=%d\n",commentSize);
    if (debug) printf("minCommentSize=%d\n",minCommentSize);
    if (debug) printf("allow_user_replay=%d\n",allow_user_replay);
    
    return 1; //happy read
    
}

int 
createSlog (void) {
    /*Creates directories, sets permissions, and opens log file */
    /* log name format in fname*/
    char hname[MAXHOSTNAMELEN];
    (void) gethostname(hname,sizeof(hname));
    //get user
    //ruser=getpwuid(geteuid())
    ruser=getlogin(); 
    if (ruser == NULL ) { //getlogin failed, try another method to get user ID
        ruser=getenv("LOGNAME"); //sudo sets LOGNAME
            if (ruser == NULL ) {
                printf("Unable to determine user login id\n");
                exit(EXIT_FAILURE);
        }
    }
    time_t tvec;
    char ldate[BUFSIZ];
    tvec = time((time_t *)NULL);
    if (debug) {
        printf ("set euid: %d\n",euid); //suid user
        printf ("real ruid: %d\n",ruid); //real user
    }

    //get the ruid passwd entry for the process
    upasswd=getpwuid((uid_t)ruid);
    if (!upasswd) {
        printf("Unknown euid, exiting\n");
        exit(EXIT_FAILURE);
    }

    if (! (upasswd->pw_shell && upasswd->pw_name && upasswd->pw_name[0] 
              && upasswd->pw_dir && upasswd->pw_dir[0] && upasswd->pw_passwd)) {
        printf("Bad password entry for uid:%d, exiting\n",ruid);
        exit(EXIT_FAILURE);
    }

    shell=upasswd->pw_shell;
    if (debug) printf("DEBUG: ruid:%s shell:%s\n",upasswd->pw_name,shell);
    if (shell == NULL)
        shell = _PATH_BSHELL;
    setenv("SHELL",shell,1);

    /* Only open fscript if we are going to write data to it.
      1) if cflg - we have a cmd to execute
         if log_all_cmds==1 -> open the file
         else return 
      2) interactive shell -> open the file
    */
    if (cflg) {
        if  (log_all_cmds==0)  {// do not log cmd output
        if (debug) printf("Logging cmd output: off\n");
        return 1;
        }
    }   
    if (debug) printf("Logging cmd output: on\n");
    // create and use LOG_DIR/sl/<year>/<month>/$sfname
    char year[10];
    char month[10];
    char day[10];
    char time[30];
    //Get formats for Year, month, day, time
    my_strftime(year, sizeof(year), "%Y", localtime(&tvec));
    my_strftime(month, sizeof(month), "%m", localtime(&tvec));
    my_strftime(day, sizeof(day), "%d", localtime(&tvec));
    my_strftime(time, sizeof(time), "%H:%M:%S",localtime(&tvec));
    
    //Create the directory structure
    const char *log_dir=STRINGIFY(LOG_DIR);
    char *yeardir;
    char *monthdir;
    asprintf(&yeardir,"%s/%s",log_dir,year);
    asprintf(&monthdir,"%s/%s",yeardir,month);

    /* Verify we can write to yearpath */
    mode_t  mode = CREATE_DIR_MODE;
    mode_t old_umask = umask (002); /* allow group write bit for mkdir */
    

    do_setuid();
    do_setgid();
    int e=0; //err number
    if (euidaccess(yeardir,W_OK)==-1) {
        // cant write to yeardir, or doesn't exist 
        e=errno; //get the error
        //printf("Darn: got error e:%d ENOENT:%d\n",e,ENOENT);
        if (e == ENOENT ) { //path doesn't exist, lets create
            if (mkdir(yeardir,mode) != 0) {
                fprintf(stderr,"Error: unable to create directory: %s\n",yeardir);
                fprintf(stderr,"Verify it exists and owned by user:group %s:%s\n",STRINGIFY(DEF_USER),STRINGIFY(DEF_GROUP));
                undo_setuid();
                undo_setgid();
                exit(EXIT_FAILURE);
            }
        }
        else { //cant write to yearpath
                
                fprintf(stderr,"Error: unable to write to directory: %s\n",yeardir);
                fprintf(stderr,"Verify it exists and owned by user:group %s:%s\n",STRINGIFY(DEF_USER),STRINGIFY(DEF_GROUP));

                if (debug) {
                    printf ("set euid: %d\n",euid); //suid user
                    printf ("real ruid: %d\n",ruid); //real user
                }

                undo_setuid();
                undo_setgid();
                exit(EXIT_FAILURE);
        }
    } //end if we can write to yearpath
    /* verify we can write to monthpath */
    e=0;
    if (euidaccess(monthdir,W_OK)==-1) {
        //cant write to monthpath, or doesn't exist...
        e=errno; //get the error
        if (e==ENOENT ) { //path doesn't exist, lets create
            if (mkdir(monthdir,mode) != 0) {
                fprintf(stderr,"Error: unable to create directory: %s\n",monthdir);
                fprintf(stderr,"Verify it exists and owned by user:group %s:%s\n",STRINGIFY(DEF_USER),STRINGIFY(DEF_GROUP));
                undo_setuid();
                undo_setgid();
                exit(EXIT_FAILURE);
            }
        }
        else { //cant write to monthpath
            fprintf(stderr,"Error: unable to write to directory: %s\n",monthdir);
            fprintf(stderr,"Verify it exists and owned by user:group %s:%s\n",STRINGIFY(DEF_USER),STRINGIFY(DEF_GROUP));
            undo_setuid();
            undo_setgid();
            exit(EXIT_FAILURE);
        }
    } //end if we can write to monthpath
    /* reset the umask */
    umask(old_umask);

    //FileName formatting 
    my_strftime(ldate, sizeof ldate, "%Y.%m.%d-%H:%M:%S.log", localtime(&tvec));
    asprintf(&fname,"%s/slog-%s-%s-%s-%s",monthdir,hname,ruser,upasswd->pw_name,ldate);
    asprintf(&sfname,"slog-%s-%s-%s-%s",hname,ruser,upasswd->pw_name,ldate);
    /* end log name format */


    /* undo setuid only if we want users to be able to read their own files */
    if (allow_user_replay==1) 
        undo_setuid();
        /* this sets the uid to the users uid, not the euid */
        
    fscript = fopen(fname, "w"); //TODO: make this a read/write, and rewind the file when done
    if (fscript == NULL) {
        undo_setuid();
        undo_setgid();
        perror(fname);
        fprintf(stderr,"Verify directory exists and owned by user:group %s:%s\n",STRINGIFY(DEF_USER),STRINGIFY(DEF_GROUP));
        fail();
    }
    /* set read only to user and group */
    if (fchmod(fileno(fscript),0440)!=0)
        fprintf(stderr,"Warning: unable to change permissions on %s\n  Session log may not be secure\n",fname);
    /* TODO: move above file create mode to config.h or ssloger.conf */
    undo_setuid();
    undo_setgid();
    return 1;
}


/*
* script -t prints time delays as floating point numbers
* The example program (scriptreplay) that we provide to handle this
* timing output is a perl script, and does not handle numbers in
* locale format (not even when "use locale;" is added).
* So, since these numbers are not for human consumption, it seems
* easiest to set LC_NUMERIC here.
*/

int
main(int argc, char **argv) {
    /* unset suid bits */
    ruid = getuid ();
    gruid = getgid (); 
    euid = geteuid ();
    geuid = getegid ();
    undo_setuid ();
    undo_setgid();

    extern int optind;
    char *p;
    int ch;
 
    progname = argv[0];
    if ((p = strrchr(progname, '/')) != NULL)
        progname = p+1;


    setlocale(LC_ALL, "");
    setlocale(LC_NUMERIC, "C");	/* see comment above */
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);

    if (argc == 2) {
        if (!strcmp(argv[1], "-V") || !strcmp(argv[1], "--version")) {
            printf(_("%s (%s)\n"),
                progname, VERSION);
            return 0;
        }
    }

    while ((ch = getopt(argc, argv, "c:fqthd")) != -1)
        switch((char)ch) {
        case 'c':
            cflg = optarg;
            break;
        case 'f':
            fflg++;
            break;
        case 'q':
            qflg++;
            break;
        case 't':
            tflg++;
            break;
        case 'h': 
            hflg=0; //default is true, cd to $HOME
            break; 
        case 'd' :
            debug=1;
            break;
        case '?':
        default:
            fprintf(stderr,
                _("usage: %s [-f] [-q] [-h] [-d] [-c command]\n"),progname);
            exit(1);
        }
    argc -= optind;
    argv += optind;
    readConf();
    //Turn on flushing by default
    //fflg++;
   
    /* Create log file cut from here */    
    createSlog();
    
    getmaster();
    if ((!qflg) && (!cflg) )
        printf(_("Sslogger started, file is %s\n"), fname);
    
    /* Scenario 1: execute a command and quit */
    /* if we get a cmd string, just exec the cmd */
    if (cflg) {
        if  (log_all_cmds==0)  {// do not log cmd output
            if (debug) printf("DEBUG: running execv cmd\n");
            char *shname;
            shname = strrchr(shell, '/'); //split at '/'
            if (shname)
                shname++;
            else {
                shname = shell;
            }

            //close stdin to avoid interactive shells
            fclose(stdin);            
            if (debug) printf("DEBUG: running exec: %s\n",cflg);
            /* Send syslog command */
            char *myCmd;
            asprintf(&myCmd,"invoked_cmd:\"%s\"",cflg);
            //comment below is "(null)" as we don't prompt prompt for a Reason
            if (!logCmd2(ruser,upasswd->pw_name,myCmd,sfname,"(null)")) {
                /*somting wong, cant open log for writing */
                exit(EXIT_FAILURE);
            }
            
            /* cd to a users Home dir? */
            if (hflg) {
                if ( (chdir (upasswd->pw_dir)) != 0) {
                    fprintf (stderr,"Warning: cannot change directory to %s\n", upasswd->pw_dir);
                }
            }
              
            usleep(10000);
            execl(shell, shname, "-c", cflg, NULL);
            //we should never get here
            perror(shell);
            exit(EXIT_FAILURE);
        }
        else { //log the output
            if (debug) printf("running popen exec: %s\n",cflg);
            /* Send syslog command */
            char *myCmd;
            asprintf(&myCmd,"invoked_cmd:\"%s\"",cflg);
            
            
            //comment below is "(null)" as we don't prompt prompt for a Reason
            if (!logCmd2(ruser,upasswd->pw_name,myCmd,sfname,"(null)")) {
                /*somting wong, cant open log for writing */
                exit(EXIT_FAILURE);
            }

            /* cd to a users Home dir? */
            if (hflg) {
                if ( (chdir (upasswd->pw_dir)) != 0) {
                    fprintf (stderr,"Warning: cannot change directory to %s\n", upasswd->pw_dir);
                }
            }

            //close stdin so user cant invoke interactive shell
            fclose (stdin);
            //Do popen here, read the output & set the exit code
            char *cmd2run;
            asprintf(&cmd2run,"exec %s",cflg);
            pinput=popen(cflg,"r");
            if (!pinput)
            {
                fprintf (stderr,
                        "Unable to run command\n");
                        exit(EXIT_FAILURE);
            }
            //read_from_pipe
            int c;
            while ((c = fgetc (pinput)) != EOF) {
                putchar (c);
                //if (c=='\n') fputc('\r',fscript); //tty kludge
                fputc(c,fscript);
            }
            int stat=0;
            int exitVal=0;
            stat=pclose(pinput);
            exitVal=WEXITSTATUS(stat);
            if (debug) fprintf(stderr,"Exit value for popen run cmd %d\n",exitVal);
            usleep(300000); //give other thread chance to read
            if (fscript!=NULL) fclose(fscript);
            exit(exitVal);
        }
    }

    /* Scenario 2: run a interactive session */
    if  (! isatty(fileno(stdout))) {
        perror("TTY required for interactive session\n");
        exit(EXIT_FAILURE);
    }

    if (minCommentSize<1 ) { 
        comment="(none)";
    }
    else
        {
        commentLen=0;
        while (1 && minCommentSize>0 ) {
            //if (comment!=NULL) free(comment);
            printf ("\nReason for invoking thus interactive shell for %s:\n",upasswd->pw_name);
            getline(&comment,&commentLen,stdin);
            if (debug) printf ("DEBUG: comment len=%d len=%d\n",strlen(comment),commentLen);
            if ( strlen(comment)>minCommentSize) break;
            printf ("Minimum required comment length is: %d\n", minCommentSize);
        }
        fputs("\n",stdout);
        
        commentLen=strlen(comment);
        if (commentLen>0)  //whack off return char
            comment[commentLen-1]='\0';
        /*Below is the max number of chars allowed for the comment*/
        /*  comment will be truncated */
        //TODO: DELETE ME if (strlen(comment) > commentSize) comment[commentSize]=(char)0;
    }

    fprintf(fscript,"Reason %s invoked interactive shell for: %s\r\n",ruser,upasswd->pw_name);
    fputs(comment,fscript);
    fputs("\r\n",fscript);
    fflush(fscript);
    fixtty();
   
    if (hflg) {
      if ( (chdir (upasswd->pw_dir)) != 0) {
          fprintf (stderr,"Warning: cannot change directory to %s\n", upasswd->pw_dir);
      }
    }

    (void) signal(SIGCHLD, finish);
    child = fork();
    if (child < 0) {
        perror("fork");
        fail();
    }
    if (child == 0) {
        subchild = child = fork();
        if (child < 0) {
            perror("fork");
            fail();
        }
        if (child) {
            dooutput();
        }
        else {
            doshell();
        }
    } else
        (void) signal(SIGWINCH, resize);
    doinput();

    return 0;
}

/* Restore the effective UID to its original value. */
void
do_setuid (void)
{
    int status;

    // user bits
    #ifdef _POSIX_SAVED_IDS
    status = seteuid (euid);
    #else
    status = setreuid (ruid, euid);
    #endif
    if (status < 0) {
        fprintf (stderr, "Couldn't set uid.\n");
        exit (status);
    }
}

/* Restore the effective GID to its original value. */
void
do_setgid (void)
{
    int status;
    //group bits
    #ifdef _POSIX_SAVED_IDS
    status = setegid (geuid);
    #else
    status = setregid (gruid, geuid);
    #endif
    if (status < 0) {
        fprintf (stderr, "Couldn't set uid.\n");
        exit (status);
    }

    
}

/* Set the effective UID to the real UID. */
void
undo_setuid (void)
{
    int status;
    //user bits
    #ifdef _POSIX_SAVED_IDS
    status = seteuid (ruid);
    #else
    status = setreuid (euid, ruid);
    #endif
    if (status < 0) {
        fprintf (stderr, "Couldn't set uid.\n");
        exit (status);
    }
}

/* Set the effective GID to the real GID. */
void
undo_setgid (void)
{
    int status;
    //group bits
    #ifdef _POSIX_SAVED_IDS
    status = setegid (gruid);
    #else
    status = setregid (geuid, gruid);
    #endif
    if (status < 0) {
        fprintf (stderr, "Couldn't set uid.\n");
        exit (status);
    }

 
}


void
doinput() {
    register int cc;
    char ibuf[BUFSIZ];
    if (fscript!=NULL) 
        (void) fclose(fscript);

    while ((cc = read(0, ibuf, BUFSIZ)) > 0)
        (void) write(master, ibuf, cc);
    done();
}

#include <sys/wait.h>

void
finish(int dummy) {
    int status;
    register int pid;
    register int die = 0;

    register int cc;
    time_t tvec;
    char obuf[BUFSIZ];
    struct timeval tv;
    double oldtime=time(NULL), newtime;
    tvec = time((time_t *)NULL);


    while ((pid = wait3(&status, WNOHANG, NULL)) > 0) {
        if (pid == child) {
            die = 1;
            //printf ("Child signal\n");
            
        }
        //printf ("Waiting\n");
    }
    if (die)
        done();
}

void
resize(int dummy) {
    /* transmit window change information to the child */
    (void) ioctl(0, TIOCGWINSZ, (char *)&win);
    (void) ioctl(slave, TIOCSWINSZ, (char *)&win);

    kill(child, SIGWINCH);
}



void
dooutput() {
    register int cc;
    time_t tvec;
    char obuf[BUFSIZ];
    struct timeval tv;
    double oldtime=time(NULL), newtime;

    (void) close(0);
#ifdef HAVE_LIBUTIL
     (void) close(slave);
#endif
    tvec = time((time_t *)NULL);
    my_strftime(obuf, sizeof obuf, "%c\r\n", localtime(&tvec));
    fprintf(fscript, _("Sslogger started on %s"), obuf);

    for (;;) {
        if (tflg)
            gettimeofday(&tv, NULL);
        cc = read(master, obuf, sizeof (obuf));
        if (cc <= 0)
            break;
        if (tflg) {
            newtime = tv.tv_sec + (double) tv.tv_usec / 1000000;
            fprintf(stderr, "%f %i\n", newtime - oldtime, cc);
            oldtime = newtime;
        }
        (void) write(1, obuf, cc);
        (void) fwrite(obuf, 1, cc, fscript);
        if (fflg)
            (void) fflush(fscript);
    }
    done();
}

void
doshell() {
    char *shname;

#if 0
    int t;

    t = open(_PATH_TTY, O_RDWR);
    if (t >= 0) {
        (void) ioctl(t, TIOCNOTTY, (char *)0);
        (void) close(t);
    }
#endif

    getslave();
    (void) close(master);
    if (fscript!=NULL) 
        (void) fclose(fscript);
    (void) dup2(slave, 0);
    (void) dup2(slave, 1);
    (void) dup2(slave, 2);
    (void) close(slave);

    shname = strrchr(shell, '/'); //split at '/'
    if (shname)
        shname++;
    else
        shname = shell;
    usleep(10000);

    /* Send  Syslog Interactive shell */
    char *myCmd;
    asprintf(&myCmd,"invoked_shell:\"%s\"",shell);
    if (!logCmd2(ruser,upasswd->pw_name,myCmd,sfname,comment)) {
        /*somting wong, cant open log for writing */
        exit(EXIT_FAILURE);
    }

    if (debug) printf ("DEBUG: execl %s - %s\n",shell, shname);
    execl(shell, "-",(char *)NULL);
    //we should never get here
    perror(shell);
    fail();
}

void
fixtty() {
    struct termios rtt;

    rtt = tt;
    cfmakeraw(&rtt);
    rtt.c_lflag &= ~ECHO;
    (void) tcsetattr(0, TCSAFLUSH, &rtt);
}

void
fail() {

    (void) kill(0, SIGTERM);
    done();
}

void
done() {
    time_t tvec;
    if (subchild) {
        if (!qflg) {
            char buf[BUFSIZ];
            tvec = time((time_t *)NULL);
            my_strftime(buf, sizeof buf, "%c\r\n", localtime(&tvec));
            fprintf(fscript, _("\nSslogger done on %s"), buf);
        }
        if (fscript!=NULL) { 
            (void) fclose(fscript);        
            /* TODO: karlos */
        }
        (void) close(master);
    } else {
        (void) tcsetattr(0, TCSAFLUSH, &tt);
        if (!qflg)
            printf(_("Sslogger done, file is %s\r\n"), fname);
    }
    exit(0);
}

void
getmaster() {
#ifdef HAVE_LIBUTIL
    (void) tcgetattr(0, &tt);
    (void) ioctl(0, TIOCGWINSZ, (char *)&win);
    if (openpty(&master, &slave, NULL, &tt, &win) < 0) {
        fprintf(stderr, _("openpty failed\n"));
        fail();
    }
#else
    char *pty, *bank, *cp;
    struct stat stb;

    pty = &line[strlen("/dev/ptyp")];
    for (bank = "pqrs"; *bank; bank++) {
        line[strlen("/dev/pty")] = *bank;
        *pty = '0';
        if (stat(line, &stb) < 0)
            break;
        for (cp = "0123456789abcdef"; *cp; cp++) {
            *pty = *cp;
            master = open(line, O_RDWR);
            if (master >= 0) {
                char *tp = &line[strlen("/dev/")];
                int ok;

                /* verify slave side is usable */
                *tp = 't';
                ok = access(line, R_OK|W_OK) == 0;
                *tp = 'p';
                if (ok) {
                    (void) tcgetattr(0, &tt);
                        (void) ioctl(0, TIOCGWINSZ,
                        (char *)&win);
                    return;
                }
                (void) close(master);
            }
        }
    }
    fprintf(stderr, _("Out of pty's\n"));
    fail();
#endif /* not HAVE_LIBUTIL */
}

void
getslave() {
#ifndef HAVE_LIBUTIL
    line[strlen("/dev/")] = 't';
    slave = open(line, O_RDWR);
    if (slave < 0) {
        perror(line);
        fail();
    }
    (void) tcsetattr(slave, TCSAFLUSH, &tt);
    (void) ioctl(slave, TIOCSWINSZ, (char *)&win);
#endif
    (void) setsid();
    (void) ioctl(slave, TIOCSCTTY, 0);
}
